#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
import logging
import sys
from distutils.util import strtobool

import simplejson as json

"""
    Utility module for accessing Ingestion Manifest
"""

LOG = logging.getLogger("ingestion")

INGESTION_MANIFEST_FILE = "ingestion-manifest.json"
CALIBRATION = "calibration"
EXECUTION_BLOCK = "execution_block"
NAASC_CALIBRATION = "naasc_calibration"
IMAGE = "image"
CONTINUUM_IMAGE = "continuum_image"
IMAGE_CUBE = "image_cube"


class IngestionManifest:
    def __init__(self, file):
        self.from_json(file)

    def to_json(self, file):
        """
        Save object in JSON file
        :param file: file to store object
        :return: None
        """
        with open(file, "w") as manifest_file:
            return json.dump(self, manifest_file, default=lambda o: o.__dict__, sort_keys=True, indent=4)

    def from_json(self, file):
        """
        Load object from JSON file
        :param file: ingestion manifest file
        :return: None
        """
        with open(file, "r") as manifest_file:
            self.__dict__ = json.load(manifest_file)

    def get_parameters(self):
        return self.parameters

    def get_telescope(self):
        return self.parameters["telescope"]

    def get_operation(self):
        return self.output_group["science_products"][0]["type"]

    def get_path(self):
        return self.parameters["ingestion_path"]

    def get_reingest(self):
        return self.parameters["reingest"]

    def get_ngas_ingest(self):
        return self.parameters["ngas_ingest"]

    def get_additional_metadata_file(self):
        if "additional_metadata" in self.__dict__["parameters"]:
            return self.get_parameters()["additional_metadata"]

    def get_collection_metadata_file(self):
        if self.has_collection_metadata_file():
            return self.get_parameters()["collection_metadata"]

    def has_collection_metadata_file(self):
        return "collection_metadata" in self.__dict__["parameters"]

    # Calibration parameters

    def get_tar_filename(self):
        return self.output_group["science_products"][0]["filename"]

    # Deprecated
    def get_execution_block_science_product_locator(self):
        return self.input_group["science_products"][0]["locator"]

    def get_execution_block_science_product_locators(self):
        # Return list of eb science product locators for ebs used in the calibration (same mous)
        return [input_science_product["locator"] for input_science_product in self.input_group["science_products"]]

    def get_input_group(self):
        if "input_group" not in self.__dict__:
            return None
        return self.input_group

    def get_output_group(self):
        return self.output_group

    def get_associate_group(self):
        if "associate_group" not in self.__dict__:
            return None
        return self.associate_group

    def get_group_science_products(self, group):
        return group["science_products"]

    def get_input_group_science_products(self):
        return self.get_input_group()["science_products"]

    def get_output_group_science_products(self):
        return self.get_output_group()["science_products"]

    def get_output_group_ancillary_products(self):
        return self.get_output_group()["ancillary_products"]

    def get_associate_group_science_products(self):
        return self.get_associate_group()["science_products"]

    def get_associate_group_locator(self):
        return self.get_associate_group_science_products()[0]["locator"]

    def get_endtime(self):
        return self.endtime

    def get_calibration_ingestion_parameters(self):
        """
        :return: path to calibration tar file, calibration tar filename,
            execution_block_science_prduct_locator, reingest, ngas_ingest
        """

        if self.get_operation() == CALIBRATION:
            return (
                self.get_path(),
                self.get_tar_filename(),
                self.get_execution_block_science_product_locators(),
                self.get_reingest(),
                self.get_ngas_ingest(),
            )
        elif self.get_operation == NAASC_CALIBRATION:
            return self.get_execution_block_science_product_locators(), self.get_reingest(), self.get_ngas_ingest()
        else:
            LOG.critical(f"Method only defined for calibration ingestion")
            sys.exit(1)

    def get_ingestion_parameters(self):
        # returning telescope, operation, input product groups, output product groups
        return (
            self.get_telescope(),
            self.get_operation(),
            self.get_input_group(),
            self.get_output_group(),
            self.get_reingest(),
            self.get_ngas_ingest(),
            self.get_path(),
        )

    def get_ingestion_parameters2(self):
        # returning telescope, operation, reingest, ngas_ingest, path
        return (
            self.get_telescope(),
            self.get_operation(),
            self.get_reingest(),
            self.get_ngas_ingest(),
            self.get_path(),
            self.get_additional_metadata_file(),
            self.get_collection_metadata_file(),
        )


def test1():
    ingestion_manifest = IngestionManifest(INGESTION_MANIFEST_FILE)
    print(ingestion_manifest.get_input_group())
    print(ingestion_manifest.get_output_group())

    for science_product in ingestion_manifest.get_input_group_science_products():
        print(f"Type: {science_product['type']} Locator: {science_product['locator']}")

    for science_product in ingestion_manifest.get_output_group_science_products():
        print(f"Type: {science_product['type']} File: {science_product['filename']}")

    for ancillary_product in ingestion_manifest.get_output_group_ancillary_products():
        print(f"Type: {ancillary_product['type']} File: {ancillary_product['filename']}")


if __name__ == "__main__":
    test1()
